@@ -151,15 +151,83 @@ notify_data_free (NotifyData *data)
 }
 
 struct _GUPnPServiceAction {
-        const char  *name;
+        GUPnPContext *context;
 
-        SoupMessage *msg;
+        char         *name;
 
-        xmlNode     *node;
+        SoupMessage  *msg;
 
-        GString     *response_str;
+        xmlDoc       *doc;
+        xmlNode      *node;
+
+        GString      *response_str;
 };
 
+GType
+gupnp_service_action_get_type (void)
+{
+        static GType our_type = 0;
+
+        if (our_type == 0)
+                our_type = g_pointer_type_register_static
+                                        ("GUPnPServiceAction");
+
+        return our_type;
+}
+
+static void
+finalize_action (GUPnPServiceAction *action)
+{
+        SoupServer *server;
+        char *response_body;
+
+        /* Embed action->response_str in a SOAP document */
+        g_string_prepend (action->response_str,
+                          "<?xml version=\"1.0\"?>"
+                          "<s:Envelope xmlns:s="
+                                "\"http://schemas.xmlsoap.org/soap/envelope/\" "
+                          "s:encodingStyle="
+                                "\"http://schemas.xmlsoap.org/soap/encoding/\">"
+                          "<s:Body>");
+
+        if (action->msg->status_code != SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+                g_string_append (action->response_str, "</u:");
+                g_string_append (action->response_str, action->name);
+                g_string_append (action->response_str, "Response>");
+        }
+
+        g_string_append (action->response_str,
+                         "</s:Body>"
+                         "</s:Envelope>");
+
+        response_body = g_string_free (action->response_str, FALSE);
+
+        soup_message_set_response (action->msg,
+                                   "text/xml",
+                                   SOUP_MEMORY_TAKE,
+                                   response_body,
+                                   strlen (response_body));
+
+        /* Server header on response */
+        soup_message_headers_append
+                        (action->msg->response_headers,
+                         "Server",
+                         gssdp_client_get_server_id
+                                (GSSDP_CLIENT (action->context)));
+
+        /* Tell soup server that response is now ready */
+        server = gupnp_context_get_server (action->context);
+        soup_server_unpause_message (server, action->msg);
+
+        /* Cleanup */
+        g_free (action->name);
+        g_object_unref (action->msg);
+        g_object_unref (action->context);
+        xmlFreeDoc (action->doc);
+
+        g_slice_free (GUPnPServiceAction, action);
+}
+
 /**
  * gupnp_service_action_get_name
  * @action: A #GUPnPServiceAction
@@ -190,14 +258,14 @@ gupnp_service_action_get_locales (GUPnPServiceAction *action)
 {
         g_return_val_if_fail (action != NULL, NULL);
 
-        return accept_language_get_locales (action->msg);
+        return http_request_get_accept_locales (action->msg);
 }
 
 /**
  * gupnp_service_action_get
  * @action: A #GUPnPServiceAction
  * @Varargs: tuples of argument name, argument type, and argument value
- * location, terminated with NULL.
+ * location, terminated with %NULL.
  *
  * Retrieves the specified action arguments.
  **/
@@ -298,7 +366,7 @@ gupnp_service_action_get_value (GUPnPServiceAction *action,
  * gupnp_service_action_set
  * @action: A #GUPnPServiceAction
  * @Varargs: tuples of return value name, return value type, and
- * actual return value, terminated with NULL.
+ * actual return value, terminated with %NULL.
  *
  * Sets the specified action return values.
  **/
@@ -403,13 +471,15 @@ gupnp_service_action_return (GUPnPServiceAction *action)
         g_return_if_fail (action != NULL);
 
         soup_message_set_status (action->msg, SOUP_STATUS_OK);
+
+        finalize_action (action);
 }
 
 /**
  * gupnp_service_action_return_error
  * @action: A #GUPnPServiceAction
  * @error_code: The error code
- * @error_description: The error description, or NULL if @error_code is
+ * @error_description: The error description, or %NULL if @error_code is
  * one of #GUPNP_CONTROL_ERROR_INVALID_ACTION,
  * #GUPNP_CONTROL_ERROR_INVALID_ARGS, #GUPNP_CONTROL_ERROR_OUT_OF_SYNC or
  * #GUPNP_CONTROL_ERROR_ACTION_FAILED, in which case a description is
@@ -484,6 +554,8 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
 
         soup_message_set_status (action->msg,
                                  SOUP_STATUS_INTERNAL_SERVER_ERROR);
+
+        finalize_action (action);
 }
 
 static void
@@ -590,7 +662,7 @@ query_state_variable (GUPnPService       *service,
 
 /* controlURL handler */
 static void
-control_server_handler (SoupServer        *soup_server,
+control_server_handler (SoupServer        *server,
                         SoupMessage       *msg, 
                         const char        *server_path,
                         GHashTable        *query,
@@ -599,12 +671,10 @@ control_server_handler (SoupServer        *soup_server,
 {
         GUPnPService *service;
         GUPnPContext *context;
-        GSSDPClient *client;
         xmlDoc *doc;
         xmlNode *action_node;
         const char *soap_action, *action_name;
         char *end;
-        gpointer response_body;
         GUPnPServiceAction *action;
 
         service = GUPNP_SERVICE (user_data);
@@ -616,7 +686,6 @@ control_server_handler (SoupServer        *soup_server,
         }
 
         context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
-        client = GSSDP_CLIENT (context);
 
         /* Get action name */
         soap_action = soup_message_headers_get (msg->request_headers,
@@ -651,64 +720,40 @@ control_server_handler (SoupServer        *soup_server,
         /* Create action structure */
         action = g_slice_new (GUPnPServiceAction);
 
-        action->name         = action_name;
-        action->msg          = msg;
+        action->name         = g_strdup (action_name);
+        action->msg          = g_object_ref (msg);
+        action->doc          = doc;
         action->node         = action_node;
         action->response_str = new_action_response_str (service, action_name);
+        action->context      = g_object_ref (context);
+
+        /* Tell soup server that response is not ready yet */
+        soup_server_pause_message (server, msg);
 
         /* QueryStateVariable? */
         if (strcmp (action_name, "QueryStateVariable") == 0)
                 query_state_variable (service, action);
         else {
-                /* Emit signal. Handler parses request and fills in response. */
-                g_signal_emit (service,
-                               signals[ACTION_INVOKED],
-                               g_quark_from_string (action_name),
-                               action);
+                GQuark action_name_quark;
 
-                /* Was it handled? */
-                if (msg->status_code == SOUP_STATUS_NONE) {
-                        /* No. */
+                action_name_quark = g_quark_from_string (action_name);
+                if (g_signal_has_handler_pending (service,
+                                                  signals[ACTION_INVOKED],
+                                                  action_name_quark,
+                                                  FALSE)) {
+                        /* Emit signal. Handler parses request and fills in
+                         * response. */
+                        g_signal_emit (service,
+                                       signals[ACTION_INVOKED],
+                                       action_name_quark,
+                                       action);
+                } else {
+                        /* No handlers attached. */
                         gupnp_service_action_return_error (action,
                                                            401,
                                                            "Invalid Action");
                 }
         }
-
-        /* Embed action->response_str in a SOAP document */
-        g_string_prepend (action->response_str, 
-                          "<?xml version=\"1.0\"?>"
-                          "<s:Envelope xmlns:s="
-                                "\"http://schemas.xmlsoap.org/soap/envelope/\" "
-                          "s:encodingStyle="
-                                "\"http://schemas.xmlsoap.org/soap/encoding/\">"
-                          "<s:Body>");
-
-        if (action->msg->status_code != SOUP_STATUS_INTERNAL_SERVER_ERROR) {
-                g_string_append (action->response_str, "</u:");
-                g_string_append (action->response_str, action_name);
-                g_string_append (action->response_str, "Response>");
-        }
-
-        g_string_append (action->response_str,
-                         "</s:Body>"
-                         "</s:Envelope>");
-
-        response_body = g_string_free (action->response_str, FALSE);
-
-        soup_message_set_response (msg,
-                                   "text/xml",
-                                   SOUP_MEMORY_TAKE,
-                                   response_body,
-                                   strlen (response_body));
-
-        /* Cleanup */
-        g_slice_free (GUPnPServiceAction, action);
-
-        /* Server header on response */
-        soup_message_headers_append (msg->response_headers,
-                                     "Server",
-                                     gssdp_client_get_server_id (client));
 }
 
 /* Generates a standard (re)subscription response */
@@ -941,7 +986,7 @@ unsubscribe (GUPnPService *service,
 
 /* eventSubscriptionURL handler */
 static void
-subscription_server_handler (SoupServer        *soup_server,
+subscription_server_handler (SoupServer        *server,
                              SoupMessage       *msg, 
                              const char        *server_path,
                              GHashTable        *query,
@@ -1280,7 +1325,8 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
          * @action: The invoked #GUPnPAction
          *
          * Emitted whenever an action is invoked. Handler should process
-         * @action.
+         * @action and must call either gupnp_service_action_return() or
+         * gupnp_service_action_return_error().
          **/
         signals[ACTION_INVOKED] =
                 g_signal_new ("action-invoked",
@@ -1293,7 +1339,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                               g_cclosure_marshal_VOID__POINTER,
                               G_TYPE_NONE,
                               1,
-                              G_TYPE_POINTER);
+                              GUPNP_TYPE_SERVICE_ACTION);
 
         /**
          * GUPnPService::query-variable
@@ -1316,7 +1362,8 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                               G_TYPE_NONE,
                               2,
                               G_TYPE_STRING,
-                              G_TYPE_POINTER);
+                              G_TYPE_POINTER /* Not G_TYPE_VALUE as this
+                                                is an outward argument! */);
 
         /**
          * GUPnPService::notify-failed
@@ -1345,7 +1392,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
  * gupnp_service_notify
  * @service: A #GUPnPService
  * @Varargs: Tuples of variable name, variable type, and variable value,
- * terminated with NULL.
+ * terminated with %NULL.
  *
  * Notifies listening clients that the properties listed in @Varargs
  * have changed to the specified values.
@@ -1367,9 +1414,9 @@ gupnp_service_notify (GUPnPService *service,
  * gupnp_service_notify_valist
  * @service: A #GUPnPService
  * @var_args: A va_list of tuples of variable name, variable type, and variable
- * value, terminated with NULL.
+ * value, terminated with %NULL.
  *
- * See gupnp_service_notify_valist(); this version takes a va_list for
+ * See gupnp_service_notify(); this version takes a va_list for
  * use by language bindings.
  **/
 void
@@ -1427,8 +1474,14 @@ notify_got_response (SoupSession *session,
         if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
                 /* Success: reset callbacks pointer */
                 data->callbacks = g_list_first (data->callbacks);
+
+        } else if (msg->status_code == SOUP_STATUS_PRECONDITION_FAILED) {
+                /* Precondition failed: Cancel subscription */
+                g_hash_table_remove (data->service->priv->subscriptions,
+                                     data->sid);
+
         } else {
-                /* Notify failed */
+                /* Other failure: Try next callback or signal failure. */
                 if (data->callbacks->next) {
                         SoupURI *uri;
                         GUPnPContext *context;
@@ -1691,6 +1744,40 @@ strip_camel_case (char *camel_str)
         return stripped;
 }
 
+static GCallback
+find_callback_by_name (GModule    *module,
+                       const char *name)
+{
+        GCallback callback;
+        char *full_name;
+
+        /* First try with 'on_' prefix */
+        full_name = g_strjoin ("_",
+                               "on",
+                               name,
+                               NULL);
+
+        if (!g_module_symbol (module,
+                              full_name,
+                              (gpointer) &callback)) {
+                g_free (full_name);
+
+                /* Now try with '_cb' postfix */
+                full_name = g_strjoin ("_",
+                                       name,
+                                       "cb",
+                                       NULL);
+
+                g_module_symbol (module,
+                                 full_name,
+                                 (gpointer) &callback);
+        }
+
+        g_free (full_name);
+
+        return callback;
+}
+
 /* Use the strings from @name_list as details to @signal_name, and connect
  * callbacks with names based on these same strings to @signal_name::string. */
 static void
@@ -1709,47 +1796,28 @@ connect_names_to_signal_handlers (GUPnPService *service,
                 GCallback callback;
                 char     *callback_name;
                 char     *signal_detail;
-                char     *tmp;
 
                 signal_detail = (char *) name_node->data;
-                tmp = strip_camel_case (signal_detail);
+                callback_name = strip_camel_case (signal_detail);
 
                 if (callback_prefix) {
-                        callback_name = g_strjoin ("_",
-                                                   callback_prefix,
-                                                   tmp,
-                                                   NULL);
-                        g_free (tmp);
-                        tmp = callback_name;
-                }
+                        char *tmp;
+
+                        tmp = g_strjoin ("_",
+                                         callback_prefix,
+                                         callback_name,
+                                         NULL);
 
-                /* First try with 'on_' prefix */
-                callback_name = g_strjoin ("_",
-                                           "on",
-                                           tmp,
-                                           NULL);
-                if (!g_module_symbol (module,
-                                      callback_name,
-                                      (gpointer) &callback)) {
                         g_free (callback_name);
-                        /* Now try with '_cb' postfix */
-                        callback_name = g_strjoin ("_",
-                                                   tmp,
-                                                   "cb",
-                                                   NULL);
-                        if (!g_module_symbol (module,
-                                              callback_name,
-                                              (gpointer) &callback)) {
-                                g_free (tmp);
-                                g_free (callback_name);
-
-                                continue;
-                        }
+                        callback_name = tmp;
                 }
 
-                g_free (tmp);
+                callback = find_callback_by_name (module, callback_name);
                 g_free (callback_name);
 
+                if (callback == NULL)
+                        continue;
+
                 signal_detail = g_strjoin ("::",
                                            signal_name,
                                            signal_detail,
@@ -1768,23 +1836,24 @@ connect_names_to_signal_handlers (GUPnPService *service,
  * gupnp_service_signals_autoconnect
  * @service: A #GUPnPService
  * @user_data: the data to pass to each of the callbacks
- * @error: return location for a GError, or NULL
+ * @error: return location for a #GError, or %NULL
  *
- * A convenience function that attempts to connect all possible "action-invoked"
- * and "query-variable" signals to appropriate callbacks for the service
- * @service. It uses service introspection and GModule's introspective
- * features. It is very simillar to glade_xml_signal_autoconnect except that it
- * attempts to guess the names of the signal handlers on its own.
+ * A convenience function that attempts to connect all possible
+ * #GUPnPService::action-invoked and #GUPnPService::query-variable signals to
+ * appropriate callbacks for the service @service. It uses service introspection
+ * and GModule's introspective features. It is very simillar to
+ * glade_xml_signal_autoconnect() except that it attempts to guess the names of
+ * the signal handlers on its own.
  *
  * For this function to do its magic, the application must name the callback
- * functions for "action-invoked" signals by striping the CamelCase off the
- * action names and either prepend "on_" or append "_cb" to them. Same goes
- * for "query-variable" signals, except that "query_" should be prepended to
- * the variable name. For example, callback function for "GetSystemUpdateID"
- * action should be either named as "get_system_update_id_cb" or
- * "on_get_system_update_id" and callback function for the query of
- * "SystemUpdateID" state variable should be named "query_system_update_id_cb"
- * or "on_query_system_update_id".
+ * functions for #GUPnPService::action-invoked signals by striping the CamelCase
+ * off the action names and either prepend "on_" or append "_cb" to them. Same
+ * goes for #GUPnPService::query-variable signals, except that "query_" should
+ * be prepended to the variable name. For example, callback function for
+ * "GetSystemUpdateID" action should be either named as
+ * "get_system_update_id_cb" or "on_get_system_update_id" and callback function
+ * for the query of "SystemUpdateID" state variable should be named
+ * "query_system_update_id_cb" or "on_query_system_update_id".
  *
  * Note that this function will not work correctly if GModule is not supported
  * on the platform or introspection is not available for service @service.
